import discord
from .Base import generate_column_types, DatabaseNotConnected
import asyncio


class EconomyAccount:
    def __init__(self, guild: int, member: int, database, table):
        self.guild = guild
        self.member = member
        self.database = database
        self.table = table

    def __str__(self):
        return f"<Account MEMBER={self.member}, GUILD={self.guild}>"

    def __lt__(self, other):
        loop = asyncio.get_event_loop()

        current_net = loop.run_until_complete(self.net())  # make another way
        other_net = loop.run_until_complete(other.net())

        return current_net < other_net

    @property
    def __checks(self):
        return EconomyManager.generate_checks(self.guild, self.member)

    async def currency(self):
        currency_data = await self.database.select(self.table, ['currency'], self.__checks)
        return currency_data["currency"]

    async def bank(self):
        bank_data = await self.database.select(self.table, ['bank'], self.__checks)
        return bank_data["bank"]

    async def net(self):
        return await self.bank() + await self.currency()

    async def change_currency(self, amount: int):
        currency = await self.currency()
        await self.database.update(self.table, {'currency': currency + amount}, self.__checks)

    async def change_bank(self, amount: int):
        bank_amount = await self.bank()
        await self.database.update(self.table, {'bank': bank_amount + amount}, self.__checks)


class EconomyManager:
    def __init__(self, bot):
        self.database = None
        self.table = None

        self.bot = bot
        self.keys = ['guild', 'member', 'currency', 'bank']

    def __check_database(self):
        if not self.database:
            raise DatabaseNotConnected(f"Database not connected."
                                       f" Connect this manager to a database using {self.__class__.__name__}")

    async def connect_to_database(self, database, table):
        types = generate_column_types(['snowflake', 'snowflake', 'snowflake', 'snowflake'],
                                      type(database.database))

        await database.create_table(table, dict(zip(self.keys, types)) if types else None, True)

        self.database = database
        self.table = table

    @staticmethod
    def generate_checks(guild: int, member: int):
        return {'guild': guild, 'member': member}

    async def create_account(self, member: discord.Member):
        self.__check_database()

        await self.database.insertifnotexists(self.table,
                                              {"guild": member.guild.id,
                                               "member": member.id,
                                               "currency": 0,
                                               "bank": 0
                                               },
                                              self.generate_checks(member.guild.id, member.id))

    async def get_account(self, member: discord.Member):
        self.__check_database()

        member_data = await self.database.select(self.table, [], self.generate_checks(member.guild.id, member.id), True)

        if member_data:
            return EconomyAccount(member.guild.id, member.id, self.database, self.table)

        return None

    async def get_leaderboard(self, guild):
        self.__check_database()

        guild_info = await self.database.select(self.table, [], {'guild': guild.id}, True)
        members = [EconomyAccount(member_info['guild'],
                                  member_info['member'],
                                  database=self.database,
                                  table=self.table) for member_info in guild_info]

        members.sort()
        return members



import aiopg
import aiosqlite
from motor import motor_asyncio
import asyncio

asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())


async def create_postgre(connection_string):
    pool = await aiopg.create_pool(connection_string)
    return await pool.acquire()


class UnsupportedDatabase(Exception):
    """Raises error when the user tries to use an unsupported database."""


class _MongoDatabase:
    def __init__(self, database):
        self.database = database

    def __str__(self):
        return f"<{self.__class__.__name__} '{self.name}'>"

    @property
    def name(self):
        return self.database.name

    async def close(self):
        self.database.client.close()

    async def insertifnotexists(self, table_name, data, checks):
        response = await self.select(table_name, [], checks, True)

        if not response:
            return await self.insert(table_name, data)

    async def insert(self, table_name, data):
        return await self.database[table_name].insert_one(data)

    async def create_table(self, table_name, columns=None, exists=False):
        if exists and table_name in await self.database.list_collection_names():
            return

        return await self.database.create_collection(table_name)

    async def update(self, table_name, data, checks):
        return await self.database[table_name].update_one(checks, {"$set": data})

    async def updateorinsert(self, table_name, data, checks, insert_data):
        response = await self.select(table_name, [], checks, True)

        if len(response) == 1:
            return await self.update(table_name, data, checks)

        return await self.insert(table_name, insert_data)

    async def delete(self, table_name, checks=None):
        return await self.database[table_name].delete_one({} if checks is None else checks)

    async def select(self, table_name, keys, checks=None, fetchall=False):
        checks = {} if checks is None else checks

        if fetchall:
            fetch = self.database[table_name].find(checks)
            result = []

            async for doc in fetch:
                current_doc = {}

                for key, value in doc.items():
                    if not keys or key in keys:
                        current_doc[key] = value

                result.append(current_doc)
        else:
            fetch = await self.database[table_name].find_one(checks)
            result = {}

            if fetch is not None:
                for key, value in fetch.items():
                    if not keys or key in keys:
                        result[key] = value
            else:
                result = None

        return result


class _SqlDatabase:
    def __str__(self):
        return f"<{self.__class__.__name__}>"

    def with_commit(func):
        async def inner(self, *args, **kwargs):
            resp = await func(self, *args, **kwargs)
            if self.commit_needed:
                await self.commit()

            return resp

        return inner

    def with_cursor(func):
        async def inner(self, *args, **kwargs):
            if self.cursor_context:
                async with self.database.cursor() as cursor:
                    resp = await func(self, cursor, *args, **kwargs)

            else:
                cursor = await self.database.cursor()
                resp = await func(self, cursor, *args, **kwargs)
                await cursor.close()

            return resp

        return inner

    def __init__(self, database):
        self.database = database
        self.place_holder = DATABASE_TYPES[type(database)]["placeholder"]
        self.cursor_context = DATABASE_TYPES[type(database)]['cursorcontext']
        self.commit_needed = DATABASE_TYPES[type(database)]['commit']

    async def commit(self):
        await self.database.commit()

    async def close(self):
        await self.database.close()

    async def insertifnotexists(self, table_name, data, checks):
        response = await self.select(table_name, [], checks, True)

        if not response:
            return await self.insert(table_name, data)

    @with_cursor
    @with_commit
    async def insert(self, cursor, table_name, data):
        query = f"INSERT INTO {table_name} ({', '.join(data.keys())}) VALUES ({', '.join([self.place_holder] * len(data.values()))})"
        await cursor.execute(query, list(data.values()))

    @with_cursor
    @with_commit
    async def create_table(self, cursor, table_name, columns=None, exists=False):
        query = f'CREATE TABLE {"IF NOT EXISTS" if exists else ""} \"{table_name}\" ('

        for column in [] if columns is None else columns:
            query += f"\n\"{column}\" {columns[column]},"
        query = query[:-1]

        query += "\n);"
        await cursor.execute(query)

    @with_cursor
    @with_commit
    async def update(self, cursor, table_name, data, checks):
        query = f"UPDATE {table_name} SET "

        if data:
            for key in data:
                query += f"{key} = {self.place_holder}, "
            query = query[:-2]

        if checks:
            query += " WHERE "
            for check in checks:
                query += f"{check} = {self.place_holder} AND "

            query = query[:-4]

        await cursor.execute(query, list(data.values()) + list(checks.values()))

    async def updateorinsert(self, table_name, data, checks, insert_data):
        response = await self.select(table_name, [], checks, True)

        if len(response) == 1:
            return await self.update(table_name, data, checks)

        return await self.insert(table_name, insert_data)

    @with_cursor
    @with_commit
    async def delete(self, cursor, table_name, checks=None):
        checks = {} if checks is None else checks

        query = f"DELETE FROM {table_name} "

        if checks:
            query += "WHERE "
            for check in checks:
                query += f"{list(check)[0]} = {self.place_holder} AND "

            query = query[:-4]

        values = [list(x.values())[0] for x in checks]
        await cursor.execute(query, values)

    @with_cursor
    async def select(self, cursor, table_name, keys, checks=None, fetchall=False):
        checks = {} if checks is None else checks

        keys = '*' if not keys else keys
        query = f"SELECT {','.join(keys)} FROM {table_name} "

        if checks:
            query += "WHERE "
            for check in checks:
                query += f"{check} = {self.place_holder} AND "

            query = query[:-4]

        await cursor.execute(query, list(checks.values()))
        columns = [x[0] for x in cursor.description]

        result = await cursor.fetchall() if fetchall else await cursor.fetchone()
        if not result:
            return result

        return [dict(zip(columns, x)) for x in result] if fetchall else dict(zip(columns, result))


DATABASE_TYPES = {
    motor_asyncio.AsyncIOMotorDatabase: {"class": _MongoDatabase, "placeholder": None},
    aiosqlite.core.Connection: {"class": _SqlDatabase, "placeholder": '?', 'cursorcontext': True, 'commit': True},
    aiopg.connection.Connection: {"class": _SqlDatabase, "placeholder": '%s', 'cursorcontext': True, 'commit': False}
}


class DatabaseManager:
    """
    A database class for easier access
    database MUST be of type sqlite3, mongodb or postgresql
    """

    @staticmethod
    def connect(database):
        if type(database) not in DATABASE_TYPES:
            raise UnsupportedDatabase(f"Database of type {type(database)} is not supported by the database manager.")

        return DATABASE_TYPES[type(database)]["class"](database)


""""
If InviteTracker is used in any way that breaks Discord TOS we, (the DiscordSuperUtils team)
are not responsible or liable in any way.
InviteTracker by DiscordSuperUtils was not intended to violate Discord TOS in any way.
In case we are contacted by Discord, we will remove any and all features that violate the Discord ToS.
Please feel free to read the Discord Terms of Service https://discord.com/terms.
"""

import discord
from discord.ext import commands


class InviteUser:
    def __init__(self, user: discord.Member, invites, ):
        self.user = user
        self.invite_list = invites

    def __str__(self):
        return f"<user={self.user.id} invites={self.invites} users_invited={self.users_invited}>"

    @property
    def invite_codes(self):
        return [invite.code for invite in self.invite_list]

    @property
    def invites(self):
        return len(self.invite_list)

    @property
    def users_invited(self):
        return sum([int(code.uses) for code in self.invite_list])


class InviteTracker:
    def __init__(self, bot: commands.Bot):
        self.bot = bot
        self.cache = {}

        self.bot.loop.create_task(self.__initialize_cache())

        self.bot.add_listener(self.__cleanup_guild_cache, 'on_guild_remove')
        self.bot.add_listener(self.__update_guild_cache, 'on_guild_add')
        self.bot.add_listener(self.__track_invite, 'on_invite_create')
        self.bot.add_listener(self.__cleanup_invite, 'on_invite_delete')

    async def get_invite(self, member: discord.Member):
        for inv in await member.guild.invites():
            for invite in self.cache[member.guild.id]:
                if invite.revoked:
                    self.cache[invite.guild.id].remove(invite)
                    return

                if invite.code == inv.code and inv.uses - invite.uses == 1:
                    await self.__update_guild_cache(member.guild)
                    return inv

    async def get_user_invites(self, member: discord.Member):
        """Returns a list of invite objects that the user created"""
        return [invite for invite in self.cache[member.guild.id] if invite.inviter.id == member.id]

    async def __initialize_cache(self):
        await self.bot.wait_until_ready()

        for guild in self.bot.guilds:
            self.cache[guild.id] = await guild.invites()

    async def __update_guild_cache(self, guild: discord.Guild):
        self.cache[guild.id] = await guild.invites()

    async def __track_invite(self, invite: discord.Invite):
        self.cache[invite.guild.id].append(invite)

    async def __cleanup_invite(self, invite: discord.Invite):
        if invite in self.cache[invite.guild.id]:
            self.cache[invite.guild.id].remove(invite)

    async def __cleanup_guild_cache(self, guild: discord.Guild):
        self.cache.pop(guild.id)

    async def fetch_inviter(self, invite: discord.Invite):
        return await self.bot.fetch_user(invite.inviter.id)

    async def fetch_user_info(self, member: discord.Member):
        """Returns InviteUser Object"""
        return InviteUser(member, await self.get_user_invites(member))



import discord
import aiohttp
import requests


class AsyncWebhook:
    def __init__(self, webhook: str):
        self.webhook = webhook

    @staticmethod
    async def to_json(embed: discord.Embed):
        return embed.to_dict()

    @classmethod
    async def pick_val(cls, embed, embeds):
        if not embed:
            return [(await cls.to_json(embd)) for embd in embeds]
        return [await cls.to_json(embed)]

    @classmethod
    async def list_to_json(cls, embeds: list):
        return [(await cls.to_json(embd)) for embd in embeds]

    async def send(self, text="", embed: discord.Embed = None, embeds: list = []):
        if embeds is None:
            embeds = []
        data = {
            "content": text,
            "embeds": await self.pick_val(embed, embeds)
        }
        async with aiohttp.ClientSession() as session:
            return await session.post(self.webhook, json=data)

    async def send_embed(self, embed: discord.Embed):
        data = {
            "content": "",
            "embeds": [self.to_json(embed)]
        }
        async with aiohttp.ClientSession() as session:
            return await session.post(self.webhook, json=data)

    async def send_embeds(self, embeds: list):
        data = {
            "content": "",
            "embeds": self.list_to_json(embeds)
        }
        async with aiohttp.ClientSession() as session:
            return await session.post(self.webhook, json=data)

    async def send_raw(self, data: dict):
        async with aiohttp.ClientSession() as session:
            return await session.post(self.webhook, json=data)


class Webhook:
    def __init__(self, webhook: str):
        self.webhook = webhook

    @staticmethod
    def to_json(embed: discord.Embed):
        return embed.to_dict()

    @classmethod
    def pick_val(cls, embed, embeds):
        if not embed:
            return [(cls.to_json(embd)) for embd in embeds]
        return [cls.to_json(embed)]

    @classmethod
    def list_to_json(cls, embeds: list):
        return [(cls.to_json(embd)) for embd in embeds]

    def send(self, text="", embed: discord.Embed = None, embeds: list = []):
        if embeds is None:
            embeds = []
        data = {
            "content": text,
            "embeds": self.pick_val(embed, embeds)
        }
        return requests.post(self.webhook, json=data)

    def send_embed(self, embed: discord.Embed):
        data = {
            "content": "",
            "embeds": [self.to_json(embed)]
        }
        return requests.post(self.webhook, json=data)

    def send_embeds(self, embeds: list):
        data = {
            "content": "",
            "embeds": self.list_to_json(embeds)
        }
        return requests.post(self.webhook, json=data)

    def send_raw(self, data: dict):
        return requests.post(self.webhook, json=data)

# it mostly works but I wanna add some stuff to it before releasing to pypi!



from .Leveling import LevelingManager
from .Database import DatabaseManager, create_postgre
from .Music import *
from .Paginator import PageManager, generate_embeds
from .ReactionRoles import ReactionManager
from .Economy import EconomyManager, EconomyAccount
from .Imaging import ImageManager
from .InviteTracker import InviteTracker

__title__ = "discordSuperUtils"
__version__ = "0.1.4"
__author__ = "Koyashie07 & Adam7100"
__license__ = "MIT"



import asyncio

import aiopg
import aiosqlite
from motor import motor_asyncio

DATABASE_TYPES = {
    motor_asyncio.AsyncIOMotorDatabase: None,  # mongo does not require any columns
    aiosqlite.core.Connection: {"snowflake": "INTEGER", "string": 'TEXT', "number": "INTEGER", "smallnumber": "INTEGER"},
    aiopg.connection.Connection: {"snowflake": "bigint", "string": 'character varying', "number": "integer", "smallnumber": "smallint"}
}


class DatabaseNotConnected(Exception):
    """Raises an error when the user tries to use a method of a manager without a database connected to it."""


def generate_column_types(types, database_type):
    database_type_configuration = DATABASE_TYPES.get(database_type)

    if database_type_configuration is None:
        return

    return [database_type_configuration[x] for x in types]


class EventManager:
    def __init__(self):
        self.events = {}

    async def call_event(self, name, *args, **kwargs):
        if name in self.events:
            for event in self.events[name]:
                await event(*args, **kwargs)

    def event(self, name=None):
        def inner(func):
            self.add_event(func, name)
            return func

        return inner

    def add_event(self, func, name=None):
        name = func.__name__ if not name else name

        if not asyncio.iscoroutinefunction(func):
            raise TypeError('Listeners must be async.')

        if name in self.events:
            self.events[name].append(func)
        else:
            self.events[name] = [func]

    def remove_event(self, func, name=None):
        name = func.__name__ if not name else name

        if name in self.events:
            self.events[name].remove(func)


from .Base import EventManager, generate_column_types, DatabaseNotConnected
from .Paginator import EmojiError

database_keys = ['guild', 'message', 'role', 'emoji', 'remove_on_reaction']


class ReactionManager(EventManager):
    def __init__(self, bot):
        super().__init__()
        self.database = None
        self.table = None
        self.bot = bot

    def __check_database(self):
        if not self.database:
            raise DatabaseNotConnected(f"Database not connected."
                                       f" Connect this manager to a database using {self.__class__.__name__}")

    async def connect_to_database(self, database, table):
        types = generate_column_types(['snowflake', 'snowflake', 'snowflake', 'string', 'smallnumber'],
                                      type(database.database))

        await database.create_table(table, dict(zip(database_keys, types)) if types else None, True)

        self.database = database
        self.table = table

        self.bot.add_listener(self.__handle_reactions, "on_raw_reaction_add")
        self.bot.add_listener(self.__handle_reactions, "on_raw_reaction_remove")

    @classmethod
    def format_data(cls, data):
        return {key: value for key, value in zip(database_keys, data)}

    @staticmethod
    def get_emoji_sql(emoji):
        if not emoji.is_custom_emoji():
            return str(emoji)

        emoji_string = f"<:{emoji.name}:{emoji.id}>"
        if emoji.animated:
            emoji_string = emoji_string[:1] + 'a' + emoji_string[1:]

        return emoji_string

    async def __handle_reactions(self, payload):
        self.__check_database()

        if payload.user_id == self.bot.user.id:
            return

        database_checks = {'guild': payload.guild_id,
                           'message': payload.message_id,
                           'emoji': self.get_emoji_sql(payload.emoji)}

        reaction_role_data = await self.database.select(self.table, database_keys, database_checks)

        if not reaction_role_data:
            return

        guild = self.bot.get_guild(payload.guild_id)
        role = guild.get_role(reaction_role_data["role"])
        channel = guild.get_channel(payload.channel_id)
        message = await channel.fetch_message(payload.message_id)
        emoji = self.get_emoji_sql(payload.emoji)

        if emoji == reaction_role_data["emoji"]:
            member = payload.member if payload.member else guild.get_member(payload.user_id)

            if role is None:
                await self.call_event('on_reaction_event', guild, channel, message, member, emoji)

            else:
                if role not in member.roles:
                    await member.add_roles(role)
                elif reaction_role_data['remove_on_reaction'] == 1:
                    await member.remove_roles(role)

    async def create_reaction(self, guild, message, role, emoji, remove_on_reaction: int):
        self.__check_database()

        await self.database.insertifnotexists(self.table,
                                              dict(zip(database_keys, [
                                                  guild.id,
                                                  message.id,
                                                  role.id if role is not None else role,
                                                  emoji,
                                                  int(remove_on_reaction)
                                              ])), {'guild': guild.id, 'message': message.id, 'emoji': emoji})

        if len(emoji) > 1:
            emoji = self.bot.get_emoji(emoji)

        try:
            await message.add_reaction(emoji)
        except Exception:
            raise EmojiError("Cannot add reaction to message.")

    async def delete_reaction(self, guild, message, emoji):
        await self.database.delete(self.table, {'guild': guild.id, 'message': message.id, 'emoji': emoji})

    async def get_reactions(self, guild):
        return await self.database.select(self.table, database_keys, {'guild': guild.id}, True)



import time
import math
from .Base import EventManager, generate_column_types, DatabaseNotConnected
import asyncio


class LevelingAccount:
    def __init__(self, database, table, guild: int, member: int, rank_multiplier=1.5):
        self.database = database
        self.table = table
        self.guild = guild
        self.member = member
        self.rank_multiplier = rank_multiplier

    def __str__(self):
        return f"<Account MEMBER={self.member}, GUILD={self.guild}>"

    def __lt__(self, other):
        loop = asyncio.get_event_loop()

        current_xp = loop.run_until_complete(self.xp())  # make another way
        other_xp = loop.run_until_complete(other.xp())

        return current_xp < other_xp

    @property
    def __checks(self):
        return LevelingManager.generate_checks(self.guild, self.member)

    async def xp(self):
        xp_data = await self.database.select(self.table, ['xp'], self.__checks)
        return xp_data["xp"]

    async def level(self):
        rank_data = await self.database.select(self.table, ['rank'], self.__checks)
        return rank_data["rank"]

    async def next_level(self):
        level_up_data = await self.database.select(self.table, ['level_up'], self.__checks)
        return level_up_data["level_up"]

    async def percentage_next_level(self):
        level_up = await self.next_level()
        xp = await self.xp()
        initial_xp = await self.initial_rank_xp()

        return math.floor(abs(xp - initial_xp) / (await level_up - initial_xp) * 100)

    async def initial_rank_xp(self):
        next_level = await self.next_level()
        return 0 if next_level == 50 else await next_level / self.rank_multiplier

    async def set_xp(self, value):
        await self.database.update(self.table, {"xp": value}, self.__checks)

    async def set_level(self, value):
        await self.database.update(self.table, {"rank": value}, self.__checks)

    async def set_next_level(self, value):
        await self.database.update(self.table, {"level_up": value}, self.__checks)


class LevelingManager(EventManager):
    def __init__(self, bot, xp_on_message=5, rank_multiplier=1.5, xp_cooldown=60):
        super().__init__()
        self.database = None
        self.table = None
        self.bot = bot
        self.xp_on_message = xp_on_message
        self.rank_multiplier = rank_multiplier
        self.xp_cooldown = xp_cooldown
        self.keys = ['guild', 'member', 'rank', 'xp', 'level_up']

        self.cooldown_members = {}

    def __check_database(self):
        if not self.database:
            raise DatabaseNotConnected(f"Database not connected."
                                       f" Connect this manager to a database using {self.__class__.__name__}")

    async def connect_to_database(self, database, table):
        types = generate_column_types(['snowflake', 'snowflake', 'number', 'number', 'number'],
                                      type(database.database))

        await database.create_table(table, dict(zip(self.keys, types)) if types else None, True)

        self.database = database
        self.table = table
        self.bot.add_listener(self.__handle_experience, "on_message")

    @staticmethod
    def generate_checks(guild: int, member: int):
        return {'guild': guild, 'member': member}

    async def __handle_experience(self, message):
        self.__check_database()

        if not message.guild or message.author.bot:
            return

        if message.guild.id not in self.cooldown_members:
            self.cooldown_members[message.guild.id] = {}

        if (time.time() - self.cooldown_members[message.guild.id].get(message.author.id, 0)) >= self.xp_cooldown:
            await self.create_account(message.author)
            member_account = await self.get_account(message.author)

            await member_account.set_xp(await member_account.xp() + self.xp_on_message)
            self.cooldown_members[message.guild.id][message.author.id] = time.time()

            leveled_up = False
            while await member_account.xp() >= await member_account.next_level():
                await member_account.set_xp(await member_account.next_level() * member_account.rank_multiplier)
                await member_account.set_level(await member_account.level() + 1)
                leveled_up = True

            if leveled_up:
                await self.call_event('on_level_up', message, member_account)

    async def create_account(self, member):
        self.__check_database()

        await self.database.insertifnotexists(self.table,
                                              dict(zip(self.keys, [member.guild.id, member.id, 1, 0, 50])),
                                              self.generate_checks(member.guild.id, member.id))

    async def get_account(self, member):
        self.__check_database()

        member_data = await self.database.select(self.table, [], self.generate_checks(member.guild.id, member.id), True)

        if member_data:
            return LevelingAccount(self.database, self.table, member.guild.id, member.id, self.rank_multiplier)

        return None

    async def get_leaderboard(self, guild):
        self.__check_database()

        guild_info = await self.database.select(self.table, [], {'guild': guild.id}, True)
        members = [LevelingAccount(self.database,
                                   self.table,
                                   member_info['guild'],
                                   member_info['member'],
                                   rank_multiplier=self.rank_multiplier)
                   for member_info in guild_info]

        members.sort()
        return members
